#include <Astri_DAQ.h>
#include <sstream>
#include <iomanip>
#include <unistd.h>
#include <ctime>

#include <Astri_Redis_DAQ_Observer.h>
#include <TCP_Protocol.h>
#include <UDP_Protocol.h>
#include <File_Archiver.h>
#include <Kafka_Provider.h>
#include <Redis_Provider.h>
#include <Astri_Monitor.h>
#include <Astri_MA_Generic.h>
#include <Packet_Structure_Json.h>

#ifndef ASTRI_PACKETS_PATH
#define ASTRI_PACKETS_PATH std::string(std::getenv("HOME")).append("/BIAS/config/AstriPackets")
#endif

using namespace inaf::oasbo;
using namespace inaf::oasbo::DAQ;
using namespace inaf::oasbo::DAQ_observers;

void DAQ::AstriDAQ::printLog(std::ostream &os, std::string message) {
	time_t now = time(nullptr);
	os << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Astri DAQ]" << "\t" << message << std::endl;
}

DAQ::AstriDAQ::AstriDAQ() {
	this->receiver = Receivers::TCPProtocolBuilder().getReceiver();
	this->archiver = Archivers::FileArchiverBuilder().getArchiver();
	this->provider = Providers::RedisProviderBuilder().getProvider();
	this->monitor = PacketMonitors::AstriPacketMonitorBuilder().getMonitor();
	std::string ma_packet(
			std::string(ASTRI_PACKETS_PATH).append(
					"/AstriMA/AstriMAGeneric.json"));
	Packets::PacketStructureJson structure(ma_packet);
	this->packet = new Packets::AstriMaGeneric(structure);

	this->currentState = Status::INIT;
	firstIdle = true;
}

DAQ::AstriDAQ::~AstriDAQ() {
	if (this->currentState != Status::STOP) {
		switchState(Status::STOP);
	}
}

std::string DAQ::AstriDAQ::getStateStr(Status state) {
	switch (state) {
	case Status::INIT:
		return std::string("INIT");
	case Status::STOP:
		return std::string("STOP");
	case Status::IDLE:
		return firstIdle ? std::string("FIRST_IDLE") : std::string("IDLE");
	case Status::RUN:
		return std::string("RUN");
	case Status::READY:
		return std::string("READY");
	default:
		return std::string("NONE");
	}
}

void DAQ::AstriDAQ::stop() {
	switchState(Status::STOP);
}

int DAQ::AstriDAQ::getDaqID() {
	return this->daq_id;
}

void DAQ::AstriDAQ::setDaqID(int daq_id) {
	this->daq_id = daq_id;
	this->updateDaqIDinConf();
}

void DAQ::AstriDAQ::incrementDaqID() {
	this->daq_id += 1;
	this->updateDaqIDinConf();
}

void DAQ::AstriDAQ::updateDaqIDinConf() {
	// update run id into the configurations
	std::map<std::string, std::string> newConf;
	newConf[AstriDAQBuilder::daq_id_key] = std::to_string(this->daq_id);
	for (auto conf : this->configurations) {
		conf->insert(newConf, AstriDAQBuilder::config_target);
		conf->pushConfigToSource(AstriDAQBuilder::config_target);
	}
}

void DAQ::AstriDAQ::switchState(Status newState) {
	if (newState == STOP) {
		printLog(std::cout, "Stopping");
		this->setCurrentState(Status::STOP);
		this->runningFlag = false;

		// stop receiving
		this->receiver->closeConnectionToClient();
		if (this->receiveAndProcessThread.joinable()) {
			this->receiveAndProcessThread.join();
		}
		if (archiver->is_open()) {
			this->archiver->close();
			updateArchiverObservers();
		}

		this->checkProvThreadFlag = false;
		if (this->checkProvConn.joinable()) {
			this->checkProvConn.join();
		}
		this->provider->close();

		std::for_each(this->observers.begin(), this->observers.end(),
				[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
					ob->stop();
				});
		printLog(std::cout, "Stopped");
		return;
	}

	std::string dest;
	switch (this->currentState) {
	case Status::READY: {
		switch (newState) {
		case Status::IDLE:
			dest = computeNewDestination(newState);
			archiver->setDest(dest);
			archiver->open();
			this->setCurrentState(newState);
			updateArchiverObservers();
			break;
		default:
			break;
		}
		break;
	}
	case Status::IDLE: {
		firstIdle = false;
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			this->setCurrentState(newState);
			dest = computeNewDestination(newState);
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::RUN:
			archiver->close();
			updateArchiverObservers();
			this->setCurrentState(newState);
			dest = computeNewDestination(newState);
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::READY:
			this->setCurrentState(Status::READY);
			archiver->close();
			updateArchiverObservers();
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				printLog(std::cerr, "Unable to connect to receiver");
				if (this->currentState != Status::STOP) { // not terminate by command
					switchState(Status::STOP);
				}
				break;
			}
			incrementDaqID();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::RUN: {
		switch (newState) {
		case Status::IDLE:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			this->setCurrentState(newState);
			dest = computeNewDestination(newState);
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::RUN:
			archiver->close();
			updateArchiverObservers();
			this->incrementDaqID();
			dest = computeNewDestination(newState);
			archiver->setDest(dest);
			archiver->open();
			updateArchiverObservers();
			break;
		case Status::READY:
			this->setCurrentState(Status::READY);
			archiver->close();
			updateArchiverObservers();
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				printLog(std::cerr, "Unable to connect to receiver");
				if (this->currentState != Status::STOP) { // not terminate by command
					switchState(Status::STOP);
				}
				break;
			}
			incrementDaqID();
			monitor->reset();
			this->switchState(IDLE);
			break;
		default:
			break;
		}
		break;
	}
	case Status::INIT:
		switch (newState) {
		case Status::READY:
			std::for_each(this->observers.begin(), this->observers.end(),
					[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
						ob->start();
					}); // start observers
			if (provider->open() > 0) {
				printLog(std::cout, "Provider connected");
			}
			this->checkProvThreadFlag = true;
			this->checkProvConn = std::thread(&AstriDAQ::connectProvider, this);

			this->setCurrentState(Status::READY);
			receiver->connectToClient();
			if (!receiver->isConnectedToClient()) {
				printLog(std::cerr, "Unable to connect to receiver");
				this->switchState(STOP);
				break;
			} else {
				printLog(std::cout, "Receiver connected");
				this->switchState(IDLE);
				this->receiveAndProcessThread = std::thread(
						[this]() {
							this->runningFlag = true;
							while (this->runningFlag) {
								if (this->currentState == Status::READY) {
									std::this_thread::sleep_for(
											std::chrono::seconds(1));
								} else {
									receiveAndProcessPacket();
								}
							}
						});
			}
			break;
		default:
			std::stringstream ss;
			ss << "Switch from " << getStateStr(this->currentState) << " to "
					<< getStateStr(newState) << " not allowed";
			printLog(std::cerr, ss.str());
			break;
		}
		break;
	default:
		break;
	}
}

int DAQ::AstriDAQ::receiveAndProcessPacket() {
	int rec = receiver->receiveFromClient(*this->packet);
	if (rec == 0) {	// connection closed
		if (this->currentState != Status::STOP) { // closed by client
			printLog(std::cout, "Receiver connection closed by client");
			this->switchState(Status::READY);
			return 0;
		} else { // closed by stop
			printLog(std::cout, "Receiver connection closed by command");
			return 0;
		}
	}
	if (rec < 0 && receiver->isConnectedToClient()) { // error in data received
		return -1;
	}
	if (rec < 0 && !receiver->isConnectedToClient()) { // connection went down, reconnect
		printLog(std::cerr, "Receiver connection went down");
		this->switchState(Status::READY);
		return -1;
	}
	if (!this->packet->isRecognizedHeader()) {
		printLog(std::cerr, "Packet header not recognized");
		return -1;
	}
	this->deliverPacket();
	Packets::AstriMaGeneric *packet =
			static_cast<Packets::AstriMaGeneric*>(this->packet); // downcast
	if (packet->getType() == 15) { //NOTIF PACKET
		printLog(std::cout, "Monitor:");
		monitor->printStats();
		std::cout << "-------------------------------" << std::endl;
		unsigned short cmd = ((this->packet->getBinaryPointer()[8]) << 8)
				+ (this->packet->getBinaryPointer()[9]);
		Status newState = cmd == 43690 ? Status::RUN : Status::IDLE;

		std::stringstream ss;
		ss << "Switching from "
				<< getStateStr(this->currentState) << " to -> "
				<< getStateStr(newState);
		printLog(std::cout, ss.str());
		this->firstIdle = false;
		this->switchState(newState);
	}
	return 1;
}

void DAQ::AstriDAQ::start() {
	incrementDaqID();
	switchState(Status::READY);
}

int DAQ::AstriDAQ::deliverPacket() {
	archiver->write(*packet);
	provider->isOpen() && provider->write(*packet);
	monitor->monit(*packet);
	return 0;
}

void DAQ::AstriDAQ::updateObservers() {
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->updateAll();
			});
}

void DAQ::AstriDAQ::updateArchiverObservers() {
	std::for_each(this->observers.begin(), this->observers.end(),
			[](inaf::oasbo::DAQ_observers::BaseDAQ_Observer *ob) {
				ob->updateArchiverStats();
			});
}

int AstriDAQ::connectProvider() {
	while (checkProvThreadFlag) {
		if (!provider->isOpen()) {
			printLog(std::cerr, "Provider not connected");
			if (provider->open() > 0) {
				printLog(std::cerr, "Provider connected");
			}
		}
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}
	return 1;
}

std::string DAQ::AstriDAQ::computeNewDestination(Status state) {
	std::string target = AstriDAQBuilder::sched_block_config_target;
	std::map<std::string, std::string> actConf;
	for (auto conf : this->configurations) {
		conf->readConfigFromSource(target);
		for (auto [k, v] : conf->getConfig()) {
			actConf[k] = v;
		}
	}

	std::stringstream ss;
	int decade = this->daq_id / 10;
	ss << std::setfill('0') << std::setw(4) << std::to_string(decade) << "/";
	ss << actConf[target + "_" + AstriDAQBuilder::date_id_key] << "_MA";
	ss << std::setfill('0') << std::setw(2)
			<< actConf[target + "_" + AstriDAQBuilder::tel_id_key] << "_";
	ss << actConf[target + "_" + AstriDAQBuilder::target_id_key] << "_";
	ss << actConf[target + "_" + AstriDAQBuilder::obs_mode_key] << "_";
	ss << std::setfill('0') << std::setw(8)
			<< actConf[target + "_" + AstriDAQBuilder::run_id_key] << "_";
	std::string mode_id;
	if (state == Status::IDLE)
		mode_id = firstIdle ? "F" : "I";
	else
		mode_id = "R";

	ss << mode_id << "_";
	ss << std::setfill('0') << std::setw(6) << this->daq_id;
	return ss.str();
}

