#include <Astri_DAQ.h>

#include <TCP_Protocol.h>
#include <UDP_Protocol.h>
#include <File_Archiver.h>
#include <Kafka_Provider.h>
#include <KafkaAvro_Provider.h>
#include <Redis_Provider.h>
#include <Astri_Monitor.h>
#include <Astri_MA_Generic.h>
#include <Astri_Horn_Generic.h>
#include <Packet_Structure_Json.h>
#include <File_Receiver.h>

using namespace inaf::oasbo::DAQ;

AstriDAQBuilder::AstriDAQBuilder() {
	this->reset();
}
AstriDAQBuilder::~AstriDAQBuilder() {
	delete daq;
}
void AstriDAQBuilder::reset() {
	this->daq = new AstriDAQ();
}


AstriDAQBuilder* AstriDAQBuilder::configFrom(Configurators::BaseConfigurator &conf) {
	daq->configurations.push_back(&conf);
	std::string target = std::string(this->config_target);
	conf.readConfigFromSource(target);
	std::map<std::string, std::string> params = conf.getConfig();


// ---- PACKET --------------------------------------------------------------------------
	std::string key = target + "_" + packet_key;
	if (params.count(key) > 0) {
		if (params[key].compare(astri_ma_packet_key) == 0) {
				std::string ma_packet(
						std::string(ASTRI_PACKETS_PATH).append(
								"/AstriMA/AstriMAGeneric.json"));
				Packets::PacketStructureJson structure(ma_packet);
			daq->packet = new Packets::AstriMaGeneric(structure);
		} else if (params[key].compare(astri_horn_packet_key) == 0) {
			std::string ma_packet(
					std::string(ASTRI_PACKETS_PATH).append(
							"/AstriHorn/AstriHornGeneric.json"));
			Packets::PacketStructureJson structure(ma_packet);
			daq->packet = new Packets::AstriHornGeneric(structure);
		}

		else { // not recognized, default to astri_ma
			std::string ma_packet(
					std::string(ASTRI_PACKETS_PATH).append(
							"/AstriMA/AstriMAGeneric.json"));
			Packets::PacketStructureJson structure(ma_packet);
			daq->packet = new Packets::AstriMaGeneric(structure);
		}
	}
// ---- RECEIVER --------------------------------------------------------------------------
	key = target+"_"+receiver_key;
	if (params.count(key) > 0) {
		if (params[key].compare(tcp_prot_key) == 0){
			daq->receiver =
					Receivers::TCPProtocolBuilder().configFrom(conf)->getReceiver();
		}
		else if (params[key].compare(udp_prot_key) == 0) {
			daq->receiver =
					Receivers::UDPProtocolBuilder().configFrom(conf)->getReceiver();
		}
		else if(params[key].compare(file_rcv_key) == 0) {
			daq->receiver =
					Receivers::FileReceiverBuilder().configFrom(conf)->getReceiver();
		}
		else { // not recognized, default to tcp
			daq->receiver = Receivers::TCPProtocolBuilder().getReceiver();
		}
	}
// ---- ARCHIVER -----------------------------------------------------------------------------
	key = target+"_"+archiver_key;
	if (params.count(key) > 0) {
		if (params[key].compare(file_arch_key) == 0) {
			daq->archiver =
					Archivers::FileArchiverBuilder().configFrom(conf)->getArchiver();
		}

		else { // not recognized, default to file
			daq->archiver = Archivers::FileArchiverBuilder().getArchiver();
		}
	}
// ---- PROVIDER --------------------------------------------------------------------
	key = target+"_"+provider_key;
	if (params.count(key) > 0) {
		if (params[key].compare(kafka_prov_key) == 0) {
			daq->provider =
					Providers::KafkaProviderBuilder().configFrom(conf)->getProvider();
		} else if (params[key].compare(redis_prov_key) == 0) {
			daq->provider =
					Providers::RedisProviderBuilder().configFrom(conf)->getProvider();
		}
		else if (params[key].compare(kafkaavro_prov_key) == 0) {
					daq->provider =
							Providers::KafkaAvroProviderBuilder().configFrom(conf)->getProvider();
				}
		else { // not recognized, default to  Redis
			daq->provider = Providers::RedisProviderBuilder().getProvider();
		}
	}
// ---- MONITOR -----------------------------------------------------------------------------
	key = target + "_" + monitor_key;
	if (params.count(key) > 0) {
		if (params[key].compare(astri_simple_monitor_key) == 0) {
			daq->monitor =
					PacketMonitors::AstriPacketMonitorBuilder().configFrom(conf)->getMonitor();
		}
		else { // not recognized, default to file
			daq->monitor =
					PacketMonitors::AstriPacketMonitorBuilder().configFrom(conf)->getMonitor();
		}
	}
// ---- DAQID -------------------------------------------------------------------------------------
	key = target + "_" + daq_id_key;
	if (params.count(key) > 0)
		daq->daq_id = std::stoi(params[key]);

// ---- SCHEDULING BLOCK -----------------------------------------------------------------------------

	target = std::string(this->sched_block_config_target);
	conf.readConfigFromSource(target);
	params = conf.getConfig();

	key = target+"_"+date_id_key;
	if (params.count(key) > 0)
		daq->scheduling_block.date = params[key];

	key = target+"_"+target_id_key;
	if (params.count(key) > 0)
		daq->scheduling_block.target_id = params[key];

	key = target+"_"+obs_mode_key;
	if (params.count(key) > 0)
		daq->scheduling_block.obs_mode = params[key];

	key = target+"_"+tel_id_key;
	if (params.count(key) > 0)
		daq->scheduling_block.tel_id = std::stoi(params[key]);

	key = target+"_"+run_id_key;
	if (params.count(key) > 0)
		daq->scheduling_block.run_id = std::stoi(params[key]);

	return this;
}

AstriDAQBuilder* AstriDAQBuilder::setProtocol(Receivers::BaseReceiver *value) {
	this->daq->receiver = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setArchiver(Archivers::BaseArchiver *value) {
	this->daq->archiver = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setProvider(Providers::BaseProvider *value) {
	this->daq->provider = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setMonitor(PacketMonitors::BasePacketMonitor *value) {
	this->daq->monitor = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setPacket(PacketLib::BasePacket *value) {
	this->daq->packet = value;
	return this;
}


AstriDAQBuilder* AstriDAQBuilder::setDateId(std::string value) {
	this->daq->scheduling_block.date = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setTargetId(std::string value) {
	this->daq->scheduling_block.target_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setObsMode(std::string value) {
	this->daq->scheduling_block.obs_mode = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setTelId(int value) {
	this->daq->scheduling_block.tel_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setRunId(int value) {
	this->daq->scheduling_block.run_id = value;
	return this;
}
AstriDAQBuilder* AstriDAQBuilder::setDaqId(int value) {
	this->daq->daq_id = value;
	return this;
}

AstriDAQ* AstriDAQBuilder::getDAQ() {
	AstriDAQ *result = this->daq;
	this->reset();
	return result;
}

const std::string AstriDAQBuilder::config_target = "astridaq";

const std::string AstriDAQBuilder::receiver_key = "protocol";
const std::string AstriDAQBuilder::archiver_key = "archiver";
const std::string AstriDAQBuilder::provider_key = "provider";
const std::string AstriDAQBuilder::packet_key = "packet";
const std::string AstriDAQBuilder::monitor_key = "monitor";

const std::string AstriDAQBuilder::file_rcv_key = "file";
const std::string AstriDAQBuilder::tcp_prot_key = "tcp";
const std::string AstriDAQBuilder::udp_prot_key = "udp";
const std::string AstriDAQBuilder::file_arch_key = "file";
const std::string AstriDAQBuilder::kafka_prov_key = "kafka";
const std::string AstriDAQBuilder::kafkaavro_prov_key = "kafkaavro";
const std::string AstriDAQBuilder::redis_prov_key = "redis";
const std::string AstriDAQBuilder::astri_ma_packet_key = "astrima";
const std::string AstriDAQBuilder::astri_horn_packet_key = "astrihorn";
const std::string AstriDAQBuilder::astri_simple_monitor_key = "astrisimple";

const std::string AstriDAQBuilder::sched_block_config_target = "schedulingblock";

const std::string AstriDAQBuilder::date_id_key = "date";
const std::string AstriDAQBuilder::target_id_key = "target_id";
const std::string AstriDAQBuilder::obs_mode_key = "obs_mode";
const std::string AstriDAQBuilder::tel_id_key = "tel_id";
const std::string AstriDAQBuilder::run_id_key = "run_id";
const std::string AstriDAQBuilder::daq_id_key = "daq_id";
